/*
 * Copyright 2013 Sam Elsharif
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package info.msxlaunchers.openmsx.launcher.persistence.game;

import info.msxlaunchers.openmsx.common.Utils;
import info.msxlaunchers.openmsx.launcher.data.extra.ExtraData;
import info.msxlaunchers.openmsx.launcher.data.game.Game;
import info.msxlaunchers.openmsx.launcher.data.game.constants.Genre;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.csvreader.CsvReader;
import com.csvreader.CsvWriter;
import com.google.inject.Inject;
import com.google.inject.Singleton;
import com.google.inject.name.Named;

/**
 * Implementation of the <code>GamePersister</code> interface that persists games on and retrieves them from the local
 * hard disk. This implementation caches the retrieved games to avoid unnecessarily getting them from disk every time
 * they're needed. That's why this class was designated with Guice's Singleton annotation to keep only one instance of it
 * in the lifetime of the Guice injector (which is the lifetime of the entire application)
 * 
 * @since v1.0
 * @author Sam Elsharif
 *
 */
@Singleton
final class CSVFileGamePersister implements GamePersister
{
	private final File databasesDirectory;
	private final File backupsDirectory;

	private final String generationMSXURL;

	private final String DATABASE_EXTENSION = ".dbo";
	private final String DATABASES_DIRECTORY = "databases";
	private final String BACKUPS_DIRECTORY = "backups";
	private final String DATABASE_BACKUP_EXTENSION = ".bak";
	private final String UNDERSCORE = "_";
	private final int MAX_DATABASE_BACKUP_NUMBER = 99;
	private final int DATABASE_BACKUP_SUFFIX_LENGTH = UNDERSCORE.length()
			+ (int)(Math.log10(MAX_DATABASE_BACKUP_NUMBER)+1)
			+ DATABASE_BACKUP_EXTENSION.length();

	private Set<Game> cache = new HashSet<Game>();
	private String currentCachedDatabase = null;

	@Inject
	CSVFileGamePersister( @Named("UserDataDirectory") String currentDirectory, @Named("GenerationMSXURL") String generationMSXURL )
	{
		this.databasesDirectory = new File( currentDirectory, DATABASES_DIRECTORY );
		this.backupsDirectory = new File( currentDirectory, BACKUPS_DIRECTORY );

		//create these two directories if they don't exist
		if( !databasesDirectory.exists() )
		{
			databasesDirectory.mkdir();
		}
		if( !backupsDirectory.exists() )
		{
			backupsDirectory.mkdir();
		}

		this.generationMSXURL = generationMSXURL;
	}
	
	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.Persister#createDatabase(java.lang.String)
	 */
	@Override
	public void createDatabase( String database )
			throws DatabaseNullNameException, DatabaseAlreadyExistsException, IOException
	{
		if( Utils.isEmpty( database ) )
		{
			throw new DatabaseNullNameException();
		}
		
		File file = new File( convertDatabaseNameToDatabaseFilename( database ) );
		
		boolean success = file.createNewFile();
		if( !success )
		{
			throw new DatabaseAlreadyExistsException( database );
		}
	}

	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.game.GamePersister#deleteDatabase(java.lang.String)
	 */
	@Override
	public void deleteDatabase( String database ) throws DatabaseNotFoundException, IOException
	{
		Utils.notNull( "database", database );

		String databaseFullname = convertDatabaseNameToDatabaseFilename( database );
		File file = new File( databaseFullname );
		
		boolean success = file.delete();
		
		if( success )
		{
			if( databaseFullname.equals( currentCachedDatabase ) )
			{
				cache.clear();
			}
		}
		else
		{
			throw new DatabaseNotFoundException( database );
		}
	}

	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.game.GamePersister#backupDatabase(java.lang.String)
	 */
	@Override
	public void backupDatabase( String database )
			throws DatabaseMaxBackupReachedException, DatabaseAlreadyExistsException, DatabaseNotFoundException, IOException
	{
		Utils.notNull( "database", database );

		File[] files = backupsDirectory.listFiles(); 
		int maxBackupNumber = 0;

		for ( File file: files ) 
		{
			if( file.isFile() ) 
			{
				String filename = file.getName();
				if( filename.endsWith( DATABASE_BACKUP_EXTENSION ) &&
						filename.startsWith( database ) &&
						filename.length() == (database + DATABASE_EXTENSION).length() + DATABASE_BACKUP_SUFFIX_LENGTH )
				{
					int dbExtensionIndex = filename.lastIndexOf( DATABASE_EXTENSION );
					int backupNumber = Utils.getNumber( filename.substring( dbExtensionIndex + 5,
							dbExtensionIndex + 7 ) );
					if( backupNumber > maxBackupNumber )
					{
						maxBackupNumber = backupNumber;
					}
				}
			}
		}

		if( maxBackupNumber >= MAX_DATABASE_BACKUP_NUMBER )
		{
			throw new DatabaseMaxBackupReachedException( database );
		}
		else
		{
			File file = new File( convertDatabaseNameToDatabaseBackupFilename( database, maxBackupNumber + 1 ) );
			
			boolean success = file.createNewFile();
			
			if( !success )
			{
				//this should not happen
				throw new DatabaseAlreadyExistsException( database );
			}
		}
	}

	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.game.GamePersister#recreateDatabase(java.lang.String)
	 */
	@Override
	public void recreateDatabase( String database )
			throws DatabaseNullNameException, DatabaseAlreadyExistsException, DatabaseNotFoundException, IOException
	{
		Utils.notNull( "database", database );

		deleteDatabase( database );
		createDatabase( database );
	}

	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.game.GamePersister#getDatabases()
	 */
	@Override
	public Set<String> getDatabases()
	{
		File[] files = databasesDirectory.listFiles(); 
		Set<String> databases = new HashSet<String>();

		for ( File file: files )
		{
			if( file.isFile() ) 
			{
				String filename = file.getName();
				if( filename.endsWith( DATABASE_EXTENSION ) )
				{
					databases.add( filename.substring( 0, filename.lastIndexOf( DATABASE_EXTENSION ) ) );
				}
			}
		}

		return databases;
	}

	/* (non-Javadoc)
	 * @see info.msxlaunchers.openmsx.launcher.persistence.game.GamePersister#updateGameExtraDataInDatabases(java.util.Map)
	 */
	@Override
	public void updateGameExtraDataInDatabases( Map<String,ExtraData> extraDataMap ) throws IOException
	{
		Utils.notNull( "extraDataMap", extraDataMap );

		Set<String> databases = getDatabases();

		for( String database: databases )
		{
			boolean databaseFound = false;
			Set<Game> newGames = null;

			try
			{
				Set<Game> games = getGames( database );
				databaseFound = true;
				newGames = new HashSet<Game>();

				for( Game game: games )
				{
					String sha1 = game.getSha1Code();

					ExtraData extraData = extraDataMap.get( sha1 );

					int msxGenID = 0;
					boolean isMSX = false;
					boolean isMSX2 = false;
					boolean isMSX2Plus = false;
					boolean isTurboR = false;
					boolean isPSG = false;
					boolean isSCC = false;
					boolean isSCCI = false;
					boolean isPCM = false;
					boolean isMSXMUSIC = false;
					boolean isMSXAUDIO = false;
					boolean isMoonsound = false;
					boolean isMIDI = false;
					String screenshotSuffix = null;
					String msxGenerationURL = null;
					int genre1 = 0;
					int genre2 = 0;
					if( extraData != null )
					{
						msxGenID = extraData.getMSXGenerationsID();
						if( msxGenID > 0 )
						{
							msxGenerationURL = generationMSXURL + msxGenID;
						}
						isMSX = extraData.isMSX();
						isMSX2 = extraData.isMSX2();
						isMSX2Plus = extraData.isMSX2Plus();
						isTurboR = extraData.isTurboR();
						isPSG = extraData.isPSG();
						isSCC = extraData.isSCC();
						isSCCI = extraData.isSCCI();
						isPCM = extraData.isPCM();
						isMSXMUSIC = extraData.isMSXMUSIC();
						isMSXAUDIO = extraData.isMSXAUDIO();
						isMoonsound = extraData.isMoonsound();
						isMIDI = extraData.isMIDI();
						screenshotSuffix = extraData.getSuffix();
						genre1 = extraData.getGenre1();
						genre2 = extraData.getGenre2();
					}

					newGames.add( Game.name( game.getName() )
						.info( game.getInfo() )
						.romA( game.getRomA() )
						.romB( game.getRomB() )
						.diskA( game.getDiskA() )
						.diskB( game.getDiskB() )
						.harddisk( game.getHarddisk() )
						.extensionRom( game.getExtensionRom() )
						.tape( game.getTape() )
						.laserdisc( game.getLaserdisc() )
						.tclScript( game.getTclScript() )
						.machine( game.getMachine() )
						.sha1Code( sha1 )
						.size( game.getSize() )
						.msxGenID( msxGenID )
						.isMSX( isMSX )
						.isMSX2( isMSX2 )
						.isMSX2Plus( isMSX2Plus )
						.isTurboR( isTurboR )
						.isPSG( isPSG )
						.isSCC( isSCC )
						.isSCCI( isSCCI )
						.isPCM( isPCM )
						.isMSXMUSIC( isMSXMUSIC )
						.isMSXAUDIO( isMSXAUDIO )
						.isMoonsound( isMoonsound )
						.isMIDI( isMIDI )
						.screenshotSuffix( screenshotSuffix )
						.info( msxGenerationURL )
						.genre1( Genre.fromValue( genre1 ) )
						.genre2( Genre.fromValue( genre2 ) )
						.build() );
				}
			}
			catch( DatabaseNotFoundException e )
			{
				//simply skip - try the next one
				databaseFound = false;
			}

			if( databaseFound )
			{
				try
				{
					recreateDatabase( database );

					//save the new game set
					saveGames( newGames, database );
				}
				catch( DatabaseNullNameException dnne )
				{
					//this shouldn't happen
				}
				catch( DatabaseAlreadyExistsException daee )
				{
					//this shouldn't happen
				}
				catch( DatabaseNotFoundException dnfe )
				{
					//this shouldn't happen
				}
				catch( GameWithNullNameException gwnne )
				{
					//this shouldn't happen
				}
				catch( GameAlreadyExistsException gaee )
				{
					//this shouldn't happen
				}
			}
		}
	}

	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.Persister#getGames(java.lang.String)
	 */
	@Override
	public Set<Game> getGames( String database )
			throws DatabaseNotFoundException, IOException
	{
		Utils.notNull( "database", database );

		initializeCacheIfNeeded( convertDatabaseNameToDatabaseFilename( database ) );

		return Collections.unmodifiableSet( cache );
	}

	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.Persister#saveGame(msx.emulator.launcher.data.Game, java.lang.String)
	 */
	@Override
	public void saveGame( Game game, String database )
			throws GameWithNullNameException, GameAlreadyExistsException, DatabaseNotFoundException, IOException
	{
		Utils.notNull( "database", database );
		Utils.notNull( "game", game );

		saveGames( Collections.singleton( game ), database );
	}

	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.game.GamePersister#saveGames(java.util.Set, java.lang.String)
	 */
	@Override
	public void saveGames( Set<Game> games, String database )
			throws GameWithNullNameException, GameAlreadyExistsException, DatabaseNotFoundException, IOException
	{
		Utils.notNull( "games", games );
		Utils.notNull( "database", database );

		String databaseFile = convertDatabaseNameToDatabaseFilename( database );

		initializeCacheIfNeeded( databaseFile );

		for( Game game: games )
		{
			if( Utils.isEmpty( game.getName() ) )
			{
				throw new GameWithNullNameException();
			}
			else if( cache.contains( game ) )
			{
				//abort the whole operation
				throw new GameAlreadyExistsException( game.getName() );
			}
			else
			{
				cache.add( game );
			}
		}

		flushCacheToDisk( databaseFile );
	}

	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.Persister#updateGame(msx.emulator.launcher.data.Game, java.lang.String)
	 */
	@Override
	public void updateGame( Game oldGame, Game newGame, String database )
			throws GameWithNullNameException, GameNotFoundException, GameAlreadyExistsException, DatabaseNotFoundException, IOException
	{
		Utils.notNull( "oldGame", oldGame );
		Utils.notNull( "newGame", newGame );
		Utils.notNull( "database", database );

		if( Utils.isEmpty( newGame.getName() ) )
		{
			throw new GameWithNullNameException();
		}

		String databaseFile = convertDatabaseNameToDatabaseFilename( database );

		initializeCacheIfNeeded( databaseFile );

		boolean foundOld = cache.contains( oldGame );
		boolean canAddNew;

		if( oldGame.equals( newGame ) )
		{
			//then they have the same name
			canAddNew = true;
		}
		else
		{
			canAddNew = !cache.contains( newGame );
		}

		if( !foundOld )
		{
			throw new GameNotFoundException( newGame.getName() );
		}
		else if( !canAddNew )
		{
			throw new GameAlreadyExistsException( newGame.getName() );
		}
		else
		{
			cache.remove( oldGame );
			cache.add( newGame );
			flushCacheToDisk( databaseFile );
		}
	}

	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.Persister#deleteGame(msx.emulator.launcher.data.Game, java.lang.String)
	 */
	@Override
	public void deleteGame( Game game, String database )
			throws GameNotFoundException, DatabaseNotFoundException, IOException
	{
		Utils.notNull( "game", game );
		Utils.notNull( "database", database );

		deleteGames( Collections.singleton( game ), database );
	}
	
	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.game.GamePersister#deleteGames(java.util.Set, java.lang.String)
	 */
	@Override
	public void deleteGames( Set<Game> games, String database )
			throws GameNotFoundException, DatabaseNotFoundException, IOException
	{
		Utils.notNull( "games", games );
		Utils.notNull( "database", database );

		String databaseFile = convertDatabaseNameToDatabaseFilename( database );

		initializeCacheIfNeeded( databaseFile );

		for( Game game: games )
		{
			boolean found = cache.remove( game );

			if( !found )
			{
				throw new GameNotFoundException( game.getName() );
			}
		}

		flushCacheToDisk( databaseFile );
	}

	/* (non-Javadoc)
	 * @see msx.emulator.launcher.persistence.game.GamePersister#moveGame(msx.emulator.launcher.data.game.Game, java.lang.String, java.lang.String)
	 */
	@Override
	public Set<Game> moveGames( Set<Game> games, String oldDatabase, String newDatabase, ActionDecider actionDecider )
			throws GameNotFoundException, DatabaseNotFoundException, GameWithNullNameException, IOException
	{
		Utils.notNull( "games", games );
		Utils.notNull( "oldDatabase", oldDatabase );
		Utils.notNull( "newDatabase", newDatabase );
		Utils.notNull( "actionDecider", actionDecider );

		String oldDatabaseFile = convertDatabaseNameToDatabaseFilename( oldDatabase );
		String newDatabaseFile = convertDatabaseNameToDatabaseFilename( newDatabase );

		//the following is a set of games that will be moved successfully
		Set<Game> movedGames = new HashSet<Game>();

		//put newDatabase's games in cache since oldDatabase is already in cache
		Set<Game> newDatabaseCache = new HashSet<Game>();
		readFromDiskAndCacheData( newDatabaseFile, newDatabaseCache );

		for( Game game: games )
		{
			if( newDatabaseCache.contains( game ) )
			{
				if( !actionDecider.isYesAll() && !actionDecider.isNoAll() )
				{
					actionDecider.promptForAction( game.getName() );
				}

				if( actionDecider.isYes() || actionDecider.isYesAll() )
				{
					newDatabaseCache.remove( game );
					newDatabaseCache.add( game );
					movedGames.add( game );
					cache.remove( game );
				}
				else if( actionDecider.isNo() || actionDecider.isNoAll() )
				{
					//do nothing - skip
				}
				else if( actionDecider.isCancel() )
				{
					break;
				}
				else
				{
					//this should not happen
					throw new RuntimeException( "At least one action must be set" );
				}
			}
			else
			{
				newDatabaseCache.add( game );
				movedGames.add( game );
				cache.remove( game );
			}
		}

		//at this point persist the two caches
		flushCacheToDisk( oldDatabaseFile );
		flushCacheToDisk( newDatabaseFile, newDatabaseCache );

		return movedGames;
	}

	private void initializeCacheIfNeeded( String database )
			throws DatabaseNotFoundException, IOException
	{
		if( currentCachedDatabase == null || !currentCachedDatabase.equals( database ) )
		{
			readFromDiskAndCacheData( database );
			currentCachedDatabase = database;
		}
	}

	private void readFromDiskAndCacheData( String database )
			throws DatabaseNotFoundException, IOException
	{
		readFromDiskAndCacheData( database, cache );
	}

	private static void readFromDiskAndCacheData( String database, Set<Game> cache  )
			throws DatabaseNotFoundException, IOException
	{
		CsvReader gameRecords = null;
		try
		{
			gameRecords = new CsvReader( database );
			cache.clear();

			while ( gameRecords.readRecord() )
			{
				try
				{
					Game game = Game.name( gameRecords.get( 0 ) )
						.info( gameRecords.get( 1 ) )
						.machine( gameRecords.get( 2 ) )
						.romA( gameRecords.get( 3 ) )
						.extensionRom( gameRecords.get( 4 ) )
						.romB( gameRecords.get( 5 ) )
						.diskA( gameRecords.get( 6 ) )
						.diskB( gameRecords.get( 7 ) )
						.tape( gameRecords.get( 8 ) )
						.harddisk( gameRecords.get( 9 ) )
						.laserdisc( gameRecords.get( 10 ) )
						.tclScript( gameRecords.get( 11 ) )
						.isMSX( getBoolean( gameRecords.get( 12 ) ) )
						.isMSX2( getBoolean( gameRecords.get( 13 ) ) )
						.isMSX2Plus( getBoolean( gameRecords.get( 14 ) ) )
						.isTurboR( getBoolean( gameRecords.get( 15 ) ) )
						.isPSG( getBoolean( gameRecords.get( 16 ) ) )
						.isSCC( getBoolean( gameRecords.get( 17 ) ) )
						.isSCCI( getBoolean( gameRecords.get( 18 ) ) )
						.isPCM( getBoolean( gameRecords.get( 19 ) ) )
						.isMSXMUSIC( getBoolean( gameRecords.get( 20 ) ) )
						.isMSXAUDIO( getBoolean( gameRecords.get( 21 ) ) )
						.isMoonsound( getBoolean( gameRecords.get( 22 ) ) )
						.isMIDI( getBoolean( gameRecords.get( 23 ) ) )
						.genre1( Genre.fromValue( Utils.getNumber( gameRecords.get( 24 ) ) ) )
						.genre2( Genre.fromValue( Utils.getNumber( gameRecords.get( 25 ) ) ) )
						.msxGenID( Utils.getNumber( gameRecords.get( 26 ) ) )
						.screenshotSuffix( gameRecords.get( 27 ) )
						.sha1Code( gameRecords.get( 28 ) )
						.size( Utils.getNumber( gameRecords.get( 29 ) ) )
						.build();

					cache.add( game );
				}
				catch( IllegalArgumentException iaex )
				{
					//just skip this game
				}
			}
		}
		catch( FileNotFoundException fnfe )
		{
			throw new DatabaseNotFoundException( database );
		}
		finally
		{
			if( gameRecords != null )
			{
				gameRecords.close();
			}
		}
	}

	private void flushCacheToDisk( String database )
			throws IOException
	{
		flushCacheToDisk( database, cache );
	}

	private static void flushCacheToDisk( String database, Set<Game> games )
			throws IOException
	{
		CsvWriter csvOutput = null;
		try
		{
			//always overwrite the existing database
			csvOutput = new CsvWriter( new FileWriter( database, false ), ',' );

			for( Game game: games )
			{
				csvOutput.write( game.getName() );
				csvOutput.write( game.getInfo() );
				csvOutput.write( game.getMachine() );
				csvOutput.write( game.getRomA() );
				csvOutput.write( game.getExtensionRom() );
				csvOutput.write( game.getRomB() );
				csvOutput.write( game.getDiskA() );
				csvOutput.write( game.getDiskB() );
				csvOutput.write( game.getTape() );
				csvOutput.write( game.getHarddisk() );
				csvOutput.write( game.getLaserdisc() );
				csvOutput.write( game.getTclScript() );
				csvOutput.write( getString( game.isMSX() ) );
				csvOutput.write( getString( game.isMSX2() ) );
				csvOutput.write( getString( game.isMSX2Plus() ) );
				csvOutput.write( getString( game.isTurboR() ) );
				csvOutput.write( getString( game.isPSG() ) );
				csvOutput.write( getString( game.isSCC() ) );
				csvOutput.write( getString( game.isSCCI() ) );
				csvOutput.write( getString( game.isPCM() ) );
				csvOutput.write( getString( game.isMSXMUSIC() ) );
				csvOutput.write( getString( game.isMSXAUDIO() ) );
				csvOutput.write( getString( game.isMoonsound() ) );
				csvOutput.write( getString( game.isMIDI() ) );
				csvOutput.write( Utils.getEnumValue( game.getGenre1() ) );
				csvOutput.write( Utils.getEnumValue( game.getGenre2() ) );
				csvOutput.write( Utils.getString( game.getMsxGenID() ) );
				csvOutput.write( game.getScreenshotSuffix() );
				csvOutput.write( game.getSha1Code() );
				csvOutput.write( Utils.getString( game.getSize() ) );

				csvOutput.endRecord();
			}
		}
		finally
		{
			if( csvOutput != null )
			{
				csvOutput.close();
			}
		}
	}
	
	private static boolean getBoolean( String string )
	{
		boolean value;

		if( string.equals( "1" ) )
		{
			value = true;
		}
		else
		{
			value = false;
		}

		return value;
	}

	private static String getString( boolean value )
	{
		return value ? "1" : "0";
	}
	
	private String convertDatabaseNameToDatabaseFilename( String database )
	{
		File file = new File( databasesDirectory, database + DATABASE_EXTENSION );
		
		return file.toString();
	}

	private String convertDatabaseNameToDatabaseBackupFilename( String database, int backupNumber )
	{
		File file = new File( backupsDirectory, database + DATABASE_EXTENSION + UNDERSCORE
				+ getPaddedNumber( backupNumber ) + DATABASE_BACKUP_EXTENSION );

		return file.toString();
	}
	
	private static String getPaddedNumber( int number )
	{
		StringBuilder stringBuilder = new StringBuilder( Utils.getString( number ) );
		
		if( number < 10 )
		{
			stringBuilder.insert( 0, '0' );
		}
		
		return stringBuilder.toString();
	}
}
