/*
 * Copyright 2014 Sam Elsharif
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package info.msxlaunchers.openmsx.launcher.ui.presenter;

import info.msxlaunchers.openmsx.common.FileTypeUtils;
import info.msxlaunchers.openmsx.common.HashUtils;
import info.msxlaunchers.openmsx.common.IOUtils;
import info.msxlaunchers.openmsx.common.Utils;
import info.msxlaunchers.openmsx.extension.ExtensionLister;
import info.msxlaunchers.openmsx.launcher.data.extra.ExtraData;
import info.msxlaunchers.openmsx.launcher.data.game.Game;
import info.msxlaunchers.openmsx.launcher.data.settings.Settings;
import info.msxlaunchers.openmsx.launcher.data.settings.constants.Language;
import info.msxlaunchers.openmsx.launcher.extra.ExtraDataGetter;
import info.msxlaunchers.openmsx.launcher.persistence.game.DatabaseNotFoundException;
import info.msxlaunchers.openmsx.launcher.persistence.game.GameAlreadyExistsException;
import info.msxlaunchers.openmsx.launcher.persistence.game.GameNotFoundException;
import info.msxlaunchers.openmsx.launcher.persistence.game.GamePersister;
import info.msxlaunchers.openmsx.launcher.persistence.game.GameWithNullNameException;
import info.msxlaunchers.openmsx.launcher.starter.EmulatorStarter;
import info.msxlaunchers.openmsx.launcher.ui.view.ProfileEditingView;
import info.msxlaunchers.openmsx.machine.InvalidMachinesDirectoryException;
import info.msxlaunchers.openmsx.machine.MachineLister;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import com.google.inject.Inject;
import com.google.inject.assistedinject.Assisted;

/**
 * Implementation of <code>ProfileEditingPresenter</code>
 * 
 * @since v1.2
 * @author Sam Elsharif
 *
 */
final class ProfileEditingPresenterImpl implements ProfileEditingPresenter
{
	private final ProfileEditingView view;
	private final MainPresenter mainPresenter;
	private final GamePersister gamePersister;
	private final EmulatorStarter emulatorStarter;
	private final Map<String,ExtraData> extraDataMap;
	private final MachineLister machineLister;
	private final ExtensionLister extensionLister;
	private final Settings settings;
	private final String currentDatabase;
	private final Game currentSelectedGame;

	@Inject
	ProfileEditingPresenterImpl( ProfileEditingView view,
			MainPresenter mainPresenter,
			GamePersister gamePersister,
			EmulatorStarter emulatorStarter,
			ExtraDataGetter extraDataGetter,
			MachineLister machineLister,
			ExtensionLister extensionLister,
			@Assisted Settings settings,
			@Assisted String currentDatabase,
			@Nullable @Assisted Game currentSelectedGame ) throws IOException
	{
		this.view = Utils.notNull( "view", view );
		this.mainPresenter = Utils.notNull( "launcherPresenter", mainPresenter );
		this.gamePersister = Utils.notNull( "gamePersister", gamePersister );
		this.emulatorStarter = Utils.notNull( "emulatorStarter", emulatorStarter );
		this.extraDataMap = Utils.notNull( "extraDataGetter", extraDataGetter ).getExtraData();
		this.machineLister = Utils.notNull( "machineLister", machineLister );
		this.extensionLister = Utils.notNull( "extensionLister", extensionLister );
		this.settings = Utils.notNull( "settings", settings );
		this.currentDatabase = Utils.notNull( "currentDatabase", currentDatabase );
		this.currentSelectedGame = currentSelectedGame;
	}

	/* (non-Javadoc)
	 * @see info.msxlaunchers.openmsx.launcher.ui.presenter.ProfileEditingPresenter#onRequestAddGameScreen(info.msxlaunchers.openmsx.launcher.data.settings.constants.Language, boolean)
	 */
	@Override
	public void onRequestAddGameScreen( Language currentLanguage, boolean currentRightToLeft ) throws LauncherException
	{
		try
		{
			view.displayAddGameScreen( this, currentLanguage, machineLister.get(), extensionLister.get(), currentRightToLeft );
		}
		catch( InvalidMachinesDirectoryException imde )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_INVALID_MACHINES_DIRECTORY );
		}
		catch( IOException ioe )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_IO );
		}
	}

	/* (non-Javadoc)
	 * @see info.msxlaunchers.openmsx.launcher.ui.presenter.ProfileEditingPresenter#onRequestEditGameScreen(info.msxlaunchers.openmsx.launcher.data.settings.constants.Language, boolean)
	 */
	@Override
	public void onRequestEditGameScreen( Language currentLanguage, boolean currentRightToLeft )
			throws LauncherException
	{
		try
		{
			view.displayEditGameScreen( this,
				currentLanguage,
				machineLister.get(),
				extensionLister.get(),
				Language.isRightToLeft( currentLanguage ),
				currentSelectedGame.getName(),
				currentSelectedGame.getInfo(),
				currentSelectedGame.getMachine(),
				currentSelectedGame.getRomA(),
				currentSelectedGame.getRomB(),
				currentSelectedGame.getExtensionRom(),
				currentSelectedGame.getDiskA(),
				currentSelectedGame.getDiskB(),
				currentSelectedGame.getTape(),
				currentSelectedGame.getHarddisk(),
				currentSelectedGame.getLaserdisc(),
				currentSelectedGame.getTclScript() );
		}
		catch( InvalidMachinesDirectoryException imde )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_INVALID_MACHINES_DIRECTORY );
		}
		catch( IOException ioe )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_IO );
		}
	}

	/* (non-Javadoc)
	 * @see info.msxlaunchers.openmsx.launcher.ui.presenter.ProfileEditingPresenter#onRequestAddGameSaveAction(java.lang.String, java.lang.String, java.lang.String, boolean, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
	 */
	@Override
	public void onRequestAddGameSaveAction( String name,
			String info,
			String machine,
			boolean romsUsed,
			String romA,
			String romB,
			String extensionRom,
			String diskA,
			String diskB,
			String tape,
			String harddisk,
			String laserdisc,
			String script) throws LauncherException
	{
		String romAEffectiveValue;
		String romBEffectiveValue;
		String extensionRomEffectiveValue;

		if( romsUsed )
		{
			romAEffectiveValue = romA;
			romBEffectiveValue = romB;
			extensionRomEffectiveValue = null;
		}
		else
		{
			romAEffectiveValue = null;
			romBEffectiveValue = null;
			extensionRomEffectiveValue = extensionRom;
		}

		Game game = null;
		
		try
		{
			game = getGameObject( name, info, machine, romAEffectiveValue, romBEffectiveValue, extensionRomEffectiveValue,
					diskA, diskB, tape, harddisk, laserdisc, script );
		}
		catch( IllegalArgumentException iae )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_EMPTY_GAME_FIELDS );
		}

		if( game == null )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_EMPTY_GAME_FIELDS );
		}

		try
		{
			gamePersister.saveGame( game, currentDatabase );

			mainPresenter.onAcceptAddGameSaveAction( game );
		}
		catch ( GameWithNullNameException gwnme )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_GAME_WITH_NULL_NAME, name );
		}
		catch ( GameAlreadyExistsException gaee )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_GAME_ALREADY_EXISTS, name );
		}
		catch ( DatabaseNotFoundException dnfe )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_DATABASE_NOT_FOUND, currentDatabase );
		}
		catch ( IOException ioe )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_IO );
		}
	}

	/* (non-Javadoc)
	 * @see info.msxlaunchers.openmsx.launcher.ui.presenter.ProfileEditingPresenter#onRequestLaunchAction(java.lang.String, boolean, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
	 */
	@Override
	public void onRequestLaunchAction( String machine,
			boolean romsUsed,
			String romA,
			String romB,
			String extensionRom,
			String diskA,
			String diskB,
			String tape,
			String harddisk,
			String laserdisc,
			String script) throws LauncherException
	{
		String extensionRomEffectiveValue;
		if( romsUsed )
		{
			extensionRomEffectiveValue = null;
		}
		else
		{
			extensionRomEffectiveValue = extensionRom;
		}

		Game game = null;
		try
		{
			game = Game.machine( machine )
				.romA( romA ).romB( romB ).extensionRom( extensionRomEffectiveValue )
				.diskA( diskA ).diskB( diskB )
				.tape( tape )
				.harddisk( harddisk )
				.laserdisc( laserdisc )
				.tclScript( script )
				.build();
		}
		catch ( IllegalArgumentException iae )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_EMPTY_GAME_FIELDS );
		}

		try
		{
			emulatorStarter.start( settings, game );
		}
		catch ( IOException e )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_CANNOT_START_OPENMSX );
		}
	}

	/* (non-Javadoc)
	 * @see info.msxlaunchers.openmsx.launcher.ui.presenter.ProfileEditingPresenter#onRequestEditGameSaveAction(java.lang.String, java.lang.String, java.lang.String, java.lang.String, boolean, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
	 */
	@Override
	public void onRequestEditGameSaveAction( String oldName,
			String newName,
			String info,
			String machine,
			boolean romsUsed,
			String romA,
			String romB,
			String extensionRom,
			String diskA,
			String diskB,
			String tape,
			String harddisk,
			String laserdisc,
			String script) throws LauncherException
	{
		String romAEffectiveValue;
		String romBEffectiveValue;
		String extensionRomEffectiveValue;

		if( romsUsed )
		{
			romAEffectiveValue = romA;
			romBEffectiveValue = romB;
			extensionRomEffectiveValue = null;
		}
		else
		{
			romAEffectiveValue = null;
			romBEffectiveValue = null;
			extensionRomEffectiveValue = extensionRom;
		}

		Game newGame = null;
		try
		{
			newGame = getGameObject( newName, info, machine, romAEffectiveValue, romBEffectiveValue,
					extensionRomEffectiveValue, diskA, diskB, tape, harddisk, laserdisc, script );
		}
		catch ( IllegalArgumentException iae )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_EMPTY_GAME_FIELDS );
		}

		try
		{
			gamePersister.updateGame( currentSelectedGame, newGame, currentDatabase );

			mainPresenter.onAcceptEditGameSaveAction( oldName, newGame );
		}
		catch ( GameWithNullNameException gwnne )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_GAME_WITH_NULL_NAME );
		}
		catch ( GameNotFoundException gnfe )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_GAME_NOT_FOUND, oldName );
		}
		catch ( GameAlreadyExistsException gaee )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_GAME_ALREADY_EXISTS, newName );
		}
		catch ( DatabaseNotFoundException dnfe )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_DATABASE_NOT_FOUND, currentDatabase );
		}
		catch ( IOException ioe )
		{
			throw new LauncherException( LauncherExceptionCode.ERR_IO );
		}
	}

	private Game getGameObject( String name,
			String info,
			String machine,
			String romA,
			String romB,
			String extensionRom,
			String diskA,
			String diskB,
			String tape,
			String harddisk,
			String laserdisc,
			String script)
	{
		Game game = null;

		String mainFile = FileTypeUtils.getMainFile( romA, romB, diskA, diskB, tape, harddisk, laserdisc, script );

		if( mainFile == null )
		{
			throw new IllegalArgumentException();
		}

		int msxGenID = 0;
		boolean isMSX = false;
		boolean isMSX2 = false;
		boolean isMSX2Plus = false;
		boolean isTurboR = false;
		boolean isPSG = false;
		boolean isSCC = false;
		boolean isSCCI = false;
		boolean isPCM = false;
		boolean isMSXMUSIC = false;
		boolean isMSXAUDIO = false;
		boolean isMoonsound = false;
		boolean isMIDI = false;
		String screenshotSuffix = null;

		String sha1Code = null;
		long fileSize = 0;
		boolean notScript = isNotScript( romA, romB, diskA, diskB, tape, harddisk, laserdisc, script );

		if( notScript )
		{
			File file = new File( mainFile );
			if( file.exists() )
			{
				if( FileTypeUtils.isZIP( file ) )
				{
					ZipFile zip = null;
					InputStream inputStream = null;
	
					try
					{
						zip = new ZipFile( file );
	
						ZipEntry firstZipEntry = zip.entries().nextElement();
	
						inputStream = zip.getInputStream( firstZipEntry );
						sha1Code = HashUtils.getSHA1Code( inputStream );
						fileSize = firstZipEntry.getSize();
					}
					catch( ZipException e )
					{
						//invalid ZIP file - skip and return null
					}
					catch( IOException e )
					{
						//shouldn't happen here
					}
					finally
					{
						IOUtils.forceClose( inputStream );
						IOUtils.forceClose( zip );
					}			
				}
				else
				{
					fileSize = file.length();
					sha1Code = HashUtils.getSHA1Code( file );
				}
	
			}
	
			ExtraData extraData = extraDataMap.get( sha1Code );
			if( extraData != null )
			{
				msxGenID = extraData.getMSXGenerationsID();
				isMSX = extraData.isMSX();
				isMSX2 = extraData.isMSX();
				isMSX2Plus = extraData.isMSX();
				isTurboR = extraData.isMSX();
				isPSG = extraData.isPSG();
				isSCC = extraData.isSCC();
				isSCCI = extraData.isSCCI();
				isPCM = extraData.isPCM();
				isMSXMUSIC = extraData.isMSXMUSIC();
				isMSXAUDIO = extraData.isMSXAUDIO();
				isMoonsound = extraData.isMoonsound();
				isMIDI = extraData.isMIDI();
				screenshotSuffix = extraData.getSuffix();
			}
		}

		game = Game.name( name )
				.info( info )
				.machine( machine )
				.romA( romA )
				.romB( romB )
				.extensionRom( extensionRom )
				.diskA( diskA )
				.diskB( diskB )
				.tape( tape )
				.harddisk( harddisk )
				.laserdisc( laserdisc )
				.tclScript( script )
				.sha1Code( sha1Code )
				.size( fileSize )
				.msxGenID( msxGenID )
				.isMSX( isMSX )
				.isMSX2( isMSX2 )
				.isMSX2Plus( isMSX2Plus )
				.isTurboR( isTurboR )
				.isPSG( isPSG )
				.isSCC( isSCC )
				.isSCCI( isSCCI )
				.isPCM( isPCM )
				.isMSXMUSIC( isMSXMUSIC )
				.isMSXAUDIO( isMSXAUDIO )
				.isMoonsound( isMoonsound )
				.isMIDI( isMIDI )
				.screenshotSuffix( screenshotSuffix )
				.build();

		return game;
	}

	private static boolean isNotScript( String romA, String romB, String diskA, String diskB,
			String tape, String harddisk, String laserdisc, String script )
	{
		return !Utils.isEmpty( romA ) || !Utils.isEmpty( romB ) ||
				!Utils.isEmpty( diskA ) || !Utils.isEmpty( diskB ) ||
				!Utils.isEmpty( tape ) || !Utils.isEmpty( harddisk ) || !Utils.isEmpty( laserdisc ) ||
				Utils.isEmpty( script );
	}
}
