/*
 * Copyright 2013 Sam Elsharif
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package info.msxlaunchers.openmsx.launcher.ui.view.swing;

import info.msxlaunchers.openmsx.common.Utils;
import info.msxlaunchers.openmsx.launcher.data.settings.Settings;
import info.msxlaunchers.openmsx.launcher.data.settings.constants.Language;
import info.msxlaunchers.openmsx.launcher.ui.presenter.LauncherException;
import info.msxlaunchers.openmsx.launcher.ui.presenter.SettingsPresenter;
import info.msxlaunchers.openmsx.launcher.ui.view.swing.component.JComboBoxWithImages;
import info.msxlaunchers.openmsx.launcher.ui.view.swing.component.MessageBoxUtil;
import info.msxlaunchers.openmsx.launcher.ui.view.swing.language.LanguageDisplayFactory;

import java.awt.Component;
import java.awt.ComponentOrientation;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Map;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.GroupLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.GroupLayout.Alignment;
import javax.swing.WindowConstants;

/**
 * Settings window class
 * 
 * @since v1.0
 * @author Sam Elsharif
 *
 */
@SuppressWarnings("serial")
public class SettingsWindow  extends JDialog implements ActionListener
{
	private final SettingsPresenter presenter;
	private final Settings settings;
	private final Set<String> databases;
	private final Map<String,String> messages;
	private final String languageCode;
	private final boolean rightToLeft;
	private final Component parent;

	private JTextField openMSXFullPathTextField;
	private JButton openMSXFullPathBrowseButton;
	private JTextField openMSXMachinesFullPathTextField;
	private JButton openMSXMachinesFullPathBrowseButton;
	private JTextField screenshotsFullPathTextField;
	private JButton screenshotsFullPathBrowseButton;
	private JComboBox defaultDatabaseComboBox;
	private JComboBoxWithImages languageComboBox;
	private JButton okButton;
	private JButton cancelButton;
	
	public SettingsWindow(SettingsPresenter presenter,
			Settings settings,
			Set<String> databases,
			Language language,
			String languageCode,
			boolean rightToLeft)
	{
		this.presenter = presenter;
		this.settings = settings;
		this.databases = databases;
		this.parent = GlobalSwingContext.getIntance().getMainWindow();

		messages = LanguageDisplayFactory.getDisplayMessages( getClass(), language);
		this.languageCode = languageCode;
		this.rightToLeft = rightToLeft;
	}

	public void displaySettings()
	{
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		setTitle(messages.get("SETTINGS"));
		setModalityType(ModalityType.APPLICATION_MODAL);
		setResizable(false);
		if(rightToLeft)
		{
			setComponentOrientation(ComponentOrientation.RIGHT_TO_LEFT);
			applyComponentOrientation(ComponentOrientation.RIGHT_TO_LEFT);
		}
		setBounds(100, 100, 510, 356);
		JLabel openMSXFullPathLabel = new JLabel("openMSX");
		JLabel openMSXMachinesFullPathLabel = new JLabel(messages.get("MACHINES"));
		JLabel screenshotsFullPathLabel = new JLabel(messages.get("SCREENSHOTS"));
		JLabel defaultDatabaseLabel = new JLabel(messages.get("DATABASE"));
		JLabel languageLabel = new JLabel(messages.get("LANGUAGE"));
		if(rightToLeft)
		{
			openMSXFullPathLabel.setHorizontalAlignment(SwingConstants.LEADING);
			openMSXMachinesFullPathLabel.setHorizontalAlignment(SwingConstants.LEADING);
			screenshotsFullPathLabel.setHorizontalAlignment(SwingConstants.LEADING);
			defaultDatabaseLabel.setHorizontalAlignment(SwingConstants.LEADING);
			languageLabel.setHorizontalAlignment(SwingConstants.LEADING);			
		}
		else
		{
			openMSXFullPathLabel.setHorizontalAlignment(SwingConstants.TRAILING);
			openMSXMachinesFullPathLabel.setHorizontalAlignment(SwingConstants.TRAILING);
			screenshotsFullPathLabel.setHorizontalAlignment(SwingConstants.TRAILING);
			defaultDatabaseLabel.setHorizontalAlignment(SwingConstants.TRAILING);
			languageLabel.setHorizontalAlignment(SwingConstants.TRAILING);			
		}
		openMSXFullPathTextField = new JTextField(settings.getOpenMSXFullPath());
		openMSXFullPathTextField.setColumns(10);
		{
			openMSXFullPathBrowseButton = new JButton("...");
			openMSXFullPathBrowseButton.addActionListener(this);
		}
		openMSXMachinesFullPathTextField = new JTextField(settings.getOpenMSXMachinesFullPath());
		openMSXMachinesFullPathTextField.setColumns(10);
		{
			openMSXMachinesFullPathBrowseButton = new JButton("...");
			openMSXMachinesFullPathBrowseButton.addActionListener(this);
		}
		screenshotsFullPathTextField = new JTextField(settings.getScreenshotsFullPath());
		screenshotsFullPathTextField.setColumns(10);
		{
			screenshotsFullPathBrowseButton = new JButton("...");
			screenshotsFullPathBrowseButton.addActionListener(this);
		}
		{
			defaultDatabaseComboBox = new JComboBox(Utils.getSortedCaseInsensitiveArray(databases));
			defaultDatabaseComboBox.setSelectedItem(settings.getDefaultDatabase());
		}
		{
			Language[] languages = Language.values();
			String languageCodes[] = new String[languages.length+1];
			String localizedLanguageNames[] = new String[languages.length+1];
			ImageIcon languageImages[] = new ImageIcon[languages.length+1];
			languageCodes[0] = null;
			localizedLanguageNames[0] = messages.get("SYSTEM_DEFAULT");
			for(int ix=0; ix < languages.length; ix++)
			{
				languageCodes[ix+1] = languages[ix].toString();
				localizedLanguageNames[ix+1] = (messages.get(languages[ix].toString()));
				languageImages[ix+1] = new ImageIcon(getClass().getResource("images/flags/" +
							languages[ix].getLocaleName().toString() + ".png"));
			}
			languageComboBox = new JComboBoxWithImages(languageCodes, localizedLanguageNames, languageImages, rightToLeft);
			languageComboBox.setMaximumRowCount(languages.length+1);
			languageComboBox.setSelectedItem(messages.get(languageCode));
		}
		{
			okButton = new JButton(messages.get("OK"));
			okButton.addActionListener(this);
		}
		{
			cancelButton = new JButton(messages.get("CANCEL"));
			cancelButton.addActionListener(this);
		}
		JPanel fullPathPanel = new JPanel();
		fullPathPanel.setBorder(BorderFactory.createTitledBorder(messages.get("DIRECTORIES")));
		JPanel generalPanel = new JPanel();
		generalPanel.setBorder(BorderFactory.createTitledBorder(messages.get("GENERAL")));
		if(rightToLeft)
		{
			fullPathPanel.setComponentOrientation(ComponentOrientation.RIGHT_TO_LEFT);
			generalPanel.setComponentOrientation(ComponentOrientation.RIGHT_TO_LEFT);
		}
		GroupLayout groupLayout = new GroupLayout(getContentPane());
		groupLayout.setHorizontalGroup(
			groupLayout.createParallelGroup(Alignment.LEADING)
				.addGroup(groupLayout.createSequentialGroup()
					.addGap(11)
					.addGroup(groupLayout.createParallelGroup(Alignment.LEADING)
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(130)
							.addComponent(openMSXFullPathTextField, GroupLayout.PREFERRED_SIZE, 278, GroupLayout.PREFERRED_SIZE))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(410)
							.addComponent(false, openMSXFullPathBrowseButton, GroupLayout.PREFERRED_SIZE, 26, GroupLayout.PREFERRED_SIZE))
						.addComponent(openMSXFullPathLabel, GroupLayout.PREFERRED_SIZE, 124, GroupLayout.PREFERRED_SIZE)

					.addGroup(groupLayout.createSequentialGroup()
							.addGap(130)
							.addComponent(openMSXMachinesFullPathTextField, GroupLayout.PREFERRED_SIZE, 278, GroupLayout.PREFERRED_SIZE))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(410)
							.addComponent(false, openMSXMachinesFullPathBrowseButton, GroupLayout.PREFERRED_SIZE, 26, GroupLayout.PREFERRED_SIZE))
						.addComponent(openMSXMachinesFullPathLabel, GroupLayout.PREFERRED_SIZE, 124, GroupLayout.PREFERRED_SIZE)
					.addGroup(groupLayout.createSequentialGroup()
							.addGap(130)
							.addComponent(screenshotsFullPathTextField, GroupLayout.PREFERRED_SIZE, 278, GroupLayout.PREFERRED_SIZE))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(410)
							.addComponent(false, screenshotsFullPathBrowseButton, GroupLayout.PREFERRED_SIZE, 26, GroupLayout.PREFERRED_SIZE))
						.addComponent(screenshotsFullPathLabel, GroupLayout.PREFERRED_SIZE, 124, GroupLayout.PREFERRED_SIZE)
						.addComponent(fullPathPanel, GroupLayout.PREFERRED_SIZE, 480, GroupLayout.PREFERRED_SIZE)))
				.addGroup(groupLayout.createSequentialGroup()
					.addGap(11)
					.addGroup(groupLayout.createParallelGroup(Alignment.LEADING)
						.addGroup(groupLayout.createSequentialGroup()
							.addComponent(languageLabel, GroupLayout.PREFERRED_SIZE, 124, GroupLayout.PREFERRED_SIZE)
							.addGap(12)
							.addGroup(groupLayout.createParallelGroup(Alignment.LEADING)
								.addComponent(defaultDatabaseComboBox, GroupLayout.PREFERRED_SIZE, 200, GroupLayout.PREFERRED_SIZE)
								.addComponent(languageComboBox, GroupLayout.PREFERRED_SIZE, 200, GroupLayout.PREFERRED_SIZE)))
						.addComponent(defaultDatabaseLabel, GroupLayout.PREFERRED_SIZE, 124, GroupLayout.PREFERRED_SIZE)
						.addComponent(generalPanel, GroupLayout.PREFERRED_SIZE, 480, GroupLayout.PREFERRED_SIZE)))
				.addGroup(groupLayout.createSequentialGroup()
					.addGap(145)
					.addComponent(false, okButton, GroupLayout.PREFERRED_SIZE, MainWindow.BUTTON_DIMENSION.width, GroupLayout.PREFERRED_SIZE)
					.addGap(8)
					.addComponent(false, cancelButton, GroupLayout.PREFERRED_SIZE, MainWindow.BUTTON_DIMENSION.width, GroupLayout.PREFERRED_SIZE))
		);
		groupLayout.setVerticalGroup(
			groupLayout.createParallelGroup(Alignment.LEADING)
				.addGroup(groupLayout.createSequentialGroup()
					.addGap(11)
					.addGroup(groupLayout.createParallelGroup(Alignment.LEADING)
							.addGroup(groupLayout.createSequentialGroup()
									.addGap(29)
									.addComponent(openMSXFullPathTextField, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE))
								.addGroup(groupLayout.createSequentialGroup()
									.addGap(29)
									.addComponent(false, openMSXFullPathBrowseButton, GroupLayout.PREFERRED_SIZE, 26, GroupLayout.PREFERRED_SIZE))
								.addGroup(groupLayout.createSequentialGroup()
									.addGap(32)
									.addComponent(openMSXFullPathLabel))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(69)
							.addComponent(openMSXMachinesFullPathTextField, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(69)
							.addComponent(false, openMSXMachinesFullPathBrowseButton, GroupLayout.PREFERRED_SIZE, 26, GroupLayout.PREFERRED_SIZE))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(72)
							.addComponent(openMSXMachinesFullPathLabel))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(109)
							.addComponent(screenshotsFullPathTextField, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(109)
							.addComponent(false, screenshotsFullPathBrowseButton, GroupLayout.PREFERRED_SIZE, 26, GroupLayout.PREFERRED_SIZE))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(112)
							.addComponent(screenshotsFullPathLabel))
						.addComponent(fullPathPanel, GroupLayout.PREFERRED_SIZE, 152, GroupLayout.PREFERRED_SIZE))
					.addGap(14)
					.addGroup(groupLayout.createParallelGroup(Alignment.LEADING)
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(70)
							.addComponent(languageLabel))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(28)
							.addComponent(defaultDatabaseComboBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE)
							.addGap(14)
							.addComponent(languageComboBox, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE, GroupLayout.PREFERRED_SIZE))
						.addGroup(groupLayout.createSequentialGroup()
							.addGap(32)
							.addComponent(defaultDatabaseLabel))
						.addComponent(generalPanel, GroupLayout.PREFERRED_SIZE, 106, GroupLayout.PREFERRED_SIZE))
					.addGap(15)
					.addGroup(groupLayout.createParallelGroup(Alignment.CENTER)
						.addComponent(okButton)
						.addComponent(cancelButton)))
		);
		getContentPane().setLayout(groupLayout);

        setLocationRelativeTo(parent);
		setVisible(true);
	}

	public void actionPerformed(ActionEvent e)
	{
		Object source = e.getSource();
		if(source == okButton)
		{
			String defaultDatabase = null;
			if( defaultDatabaseComboBox.getSelectedItem() != null)
			{
				defaultDatabase = defaultDatabaseComboBox.getSelectedItem().toString();
			}
			try
			{
				presenter.onRequestSettingsAction(openMSXFullPathTextField.getText(),
						openMSXMachinesFullPathTextField.getText(),
						screenshotsFullPathTextField.getText(),
						defaultDatabase,
						languageComboBox.getSelectedCode());
				dispose();
			}
			catch(LauncherException le)
			{
				MessageBoxUtil.showErrorMessageBox(this, le, messages);
			}
		}
		else if(source == cancelButton)
		{
			dispose();
		}
		else if(source == openMSXFullPathBrowseButton)
		{
			JFileChooser chooser = new JFileChooser();
		    chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		    chooser.setAcceptAllFileFilterUsed(false);
			int returnVal = chooser.showOpenDialog(this);
			if (returnVal == JFileChooser.APPROVE_OPTION)
			{
				openMSXFullPathTextField.setText(chooser.getSelectedFile().getAbsolutePath());
			}
		}
		else if(source == openMSXMachinesFullPathBrowseButton)
		{
			JFileChooser chooser = new JFileChooser();
		    chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		    chooser.setAcceptAllFileFilterUsed(false);
			int returnVal = chooser.showOpenDialog(this);
			if (returnVal == JFileChooser.APPROVE_OPTION)
			{
				openMSXMachinesFullPathTextField.setText(chooser.getSelectedFile().getAbsolutePath());
			}
		}
		else if(source == screenshotsFullPathBrowseButton)
		{
			JFileChooser chooser = new JFileChooser();
		    chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		    chooser.setAcceptAllFileFilterUsed(false);
			int returnVal = chooser.showOpenDialog(this);
			if (returnVal == JFileChooser.APPROVE_OPTION)
			{
				screenshotsFullPathTextField.setText(chooser.getSelectedFile().getAbsolutePath());
			}
		}
	}
}
